@@ -96,6 +96,11 @@ def add(i):
               (current_repos)            - if resolving dependencies on other repos, list of repos being updated (to avoid infinite recursion)
 
               (describe)                 - describe repository for Artifact Evaluation (see http://cTuning.org/ae)
+
+              (stable)        - take stable version (highly experimental)
+              (version)       - checkout version (default - stable)
+              (branch)        - git branch
+              (checkout)      - git checkout
             }
 
     Output: {
@@ -106,7 +111,6 @@ def add(i):
             }
 
     """
-
     # Check if global writing is allowed
     r=ck.check_writing({})
     if r['return']>0: return r
@@ -118,17 +122,69 @@ def add(i):
     di=i.get('data_uid','')
     dn=i.get('data_name','')
 
+    stable=i.get('stable','')
+    version=i.get('version','')
+    if stable=='yes': version='stable'
+    branch=i.get('branch','')
+    checkout=i.get('checkout','')
+
     cr=i.get('current_repos',[])
 
     rdeps=i.get('repo_deps',[])
 
     quiet=i.get('quiet','')
 
-    zp=i.get('zip','')
     overwrite=i.get('overwrite','')
 
-    gz=i.get('gitzip','')
     ptr=''
+    zp=i.get('zip','')
+    gz=i.get('gitzip','')
+
+    if zp!='':
+       if zp.startswith('~'):
+          from os.path import expanduser
+          home = expanduser("~")
+
+          zp=os.path.abspath(home+os.sep+zp[1:])
+
+       quiet='yes'
+
+       if d=='':
+          # Try to get data UOA
+          if not os.path.isfile(zp):
+             return {'return':1, 'error':'zip file not found'}
+
+          # Try to get .ckr.json
+          import zipfile
+
+          try:
+             with zipfile.ZipFile(zp) as z:
+                if ck.cfg['repo_file'] in z.namelist():
+                   x=z.open(ck.cfg['repo_file'])
+                   y=x.read()
+
+                   r=ck.convert_json_str_to_dict({'str':y, 'skip_quote_replacement':'yes'})
+                   if r['return']>0: return r
+                   yd=r['dict']
+
+                   d=yd.get('data_uoa','')
+                   di=yd.get('data_uid','')
+                   dn=yd.get('data_name','')
+
+                   x.close()
+             z.close()
+          except Exception as e:
+             return {'return':1, 'error':'problem reading zip file ('+format(e)+')'}
+
+          if d=='':
+             x1=os.path.basename(zp)
+             d=os.path.splitext(x1)[0]
+             if d.startswith('ckr-'):
+                d=d[4:]
+
+             if o=='con':
+                ck.out('Auto-detected repo name from zip filename: '+d)
+
     if gz=='yes':
        zp=ck.cfg['default_shared_repo_url']+'/'+d+'/archive/master.zip'
        ptr=d+'-master/'
@@ -454,6 +510,9 @@ def add(i):
     r=deps({'path':p,
             'current_path':cr,
             'how':how,
+            'version':version,
+            'branch':branch,
+            'checkout':checkout,
             'out':o})
     if r['return']>0: return r
 
@@ -691,6 +750,13 @@ def pull(i):
               (current_repos) - if resolving dependencies on other repos, list of repos being updated (to avoid infinite recursion)
 
               (git)           - if 'yes', use git protocol instead of https
+
+              (ignore_pull)   - useful just for switching to another branch
+
+              (stable)        - take stable version (highly experimental)
+              (version)       - checkout version (default - stable)
+              (branch)        - git branch
+              (checkout)      - git checkout
             }
 
     Output: {
@@ -710,6 +776,14 @@ def pull(i):
     t=i.get('type','')
     url=i.get('url','')
 
+    stable=i.get('stable','')
+    version=i.get('version','')
+    if stable=='yes': version='stable'
+    branch=i.get('branch','')
+    checkout=i.get('checkout','')
+
+    ip=i.get('ignore_pull','')
+
     cr=i.get('current_repos',[])
 
     tt='pull'
@@ -760,7 +834,8 @@ def pull(i):
               if t!='':
                  p=d.get('path','')
                  url=d.get('url','')
-                 pp.append({'path':p, 'type':t, 'url':url, 'data_uoa':duoa})
+                 checkouts=d.get('checkouts',{})
+                 pp.append({'path':p, 'type':t, 'url':url, 'data_uoa':duoa, 'checkouts':checkouts})
        else:
           # Loading repo
           r=ck.access({'action':'load',
@@ -791,8 +866,9 @@ def pull(i):
           p=d['path']
           t=d.get('shared','')
           url=d.get('url','')
+          checkouts=d.get('checkouts',{})
 
-          pp.append({'path':p, 'type':t, 'url':url, 'data_uoa':duoa})
+          pp.append({'path':p, 'type':t, 'url':url, 'data_uoa':duoa, 'checkouts':checkouts})
 
     # Updating ...
     for q in pp:
@@ -836,7 +912,11 @@ def pull(i):
               return{'return':1, 'error':'git command line client is not found - please, install it or download repo as zip'}
 
            # Continue
-           px=os.getcwd()
+           try:
+              px=os.getcwd()
+           except OSError:
+              from os.path import expanduser
+              px=expanduser("~")
 
            if not os.path.isdir(p):
               os.makedirs(p)
@@ -846,23 +926,25 @@ def pull(i):
               ck.out('  cd '+p)
            os.chdir(p)
 
-           s=ck.cfg['repo_types'][t][tt].replace('$#url#$', url).replace('$#path#$', p)
+           r=0
+           if ip!='yes':
+              s=ck.cfg['repo_types'][t][tt].replace('$#url#$', url).replace('$#path#$', p)
 
-           if o=='con':
-              ck.out('  '+s)
-              ck.out('')
+              if o=='con':
+                 ck.out('  '+s)
+                 ck.out('')
 
-           r=os.system(s)
+              r=os.system(s)
 
-           if o=='con': 
-              ck.out('')
+              if o=='con': 
+                 ck.out('')
 
            os.chdir(px) # Restore path
 
            if r>0:
               if o=='con':
                  ck.out('')
-                 ck.out(' WARNING: repository update likely failed - exit code '+str(r))
+                 ck.out(' WARNING: repository update likely failed OR IN A DIFFERENT BRANCH/CHECKOUT (git exit code: '+str(r)+')')
                  ck.out('')
                  rx=ck.inp({'text': 'Would you like to continue (Y/n)?: '})
                  x=rx['string'].lower()
@@ -884,6 +966,9 @@ def pull(i):
            r=deps({'path':p,
                    'current_path':cr,
                    'how':'pull',
+                   'version':version,
+                   'branch':branch,
+                   'checkout':checkout,
                    'out':o})
            if r['return']>0: return r
 
@@ -1144,22 +1229,6 @@ def recache(i):
 
     return {'return':0}
 
-##############################################################################
-# Remove files and dirs even if read only (internal use)
-def rm_read_only(f,p,e):
-    import os
-    import stat
-    import errno
-
-    ex=e[1]
-
-    if ex.errno!=errno.EACCES or f not in (os.rmdir,os.remove): raise
-
-    os.chmod(p,stat.S_IRWXU|stat.S_IRWXG|stat.S_IRWXO)
-    f(p)
-
-    return
-
 ##############################################################################
 # Remove information about repository
 
@@ -1247,7 +1316,8 @@ def rm(i):
        if wf=='yes' and p!='':
           if o=='con': ck.out('Removing entries from the repository ...')
           import shutil
-          if os.path.isdir(p): shutil.rmtree(p, onerror=rm_read_only)
+          if os.path.isdir(p): 
+             shutil.rmtree(p, onerror=ck.rm_read_only)
 
        if o=='con': 
           ck.out('')
@@ -1378,6 +1448,12 @@ def zip(i):
 
     pfn=os.path.join(ap, an)
 
+    if pfn.startswith('~'):
+       from os.path import expanduser
+       home = expanduser("~")
+
+       pfn=os.path.abspath(home+os.sep+pfn[1:])
+
     if os.path.isfile(pfn):
        if i.get('overwrite','')=='yes':
           os.remove(pfn)
@@ -1676,6 +1752,10 @@ def deps(i):
               (current_repos) - list of repos being updated (to avoid infinite recursion)
 
               (how)           - 'pull' (default) or 'add'
+
+              (version)       - checkout version (default - stable)
+              (branch)        - git branch
+              (checkout)      - git checkout
             }
 
     Output: {
@@ -1715,16 +1795,52 @@ def deps(i):
 
           d=r['dict']
 
+          # Check checkouts
+          version=i.get('version','')
+          branch=i.get('branch','')
+          checkout=i.get('checkout','')
+
+          if version!='':
+             cx=d.get('dict',{}).get('checkouts',{}).get(version,{})
+             branch=cx.get('branch','')
+             checkout=cx.get('checkout','')
+
+          ppp=os.getcwd()
+
+          os.chdir(p)
+
+          if branch!='':
+             if o=='con':
+                ck.out('  ====================================')
+                ck.out('  git checkout '+branch)
+                ck.out('')
+
+             r=ck.run_and_get_stdout({'cmd':['git','checkout',branch]})
+             ck.out(r.get('stdout',''))
+             ck.out(r.get('stderr',''))
+
+          if checkout!='':
+             if o=='con':
+                ck.out('  ====================================')
+                ck.out('  git checkout '+checkout)
+                ck.out('')
+
+             r=ck.run_and_get_stdout({'cmd':['git','checkout',checkout]})
+             ck.out(r.get('stdout',''))
+             ck.out(r.get('stderr',''))
+
+          os.chdir(ppp)
+
           rp1=d.get('dict',{}).get('repo_deps',[])
           if len(rp1)==0:
-             rp1=d.get('repo_deps',[]) # for compatibility ...
+             rp1=d.get('repo_deps',[]) # for backwards compatibility ...
 
           rp2=[]
           rp=[]
 
           if len(rp1)>0:
              for xruoa in rp1:
-                 if type(xruoa)!=list:
+                 if type(xruoa)!=list: # for backwards compatibility
                     ruoa=xruoa.get('repo_uoa','')
                     if xruoa.get('repo_uid','')!='': ruoa=xruoa['repo_uid']
                     if ruoa!='' and ruoa not in cr:
@@ -1751,7 +1867,18 @@ def deps(i):
                        rp.append(xruoa)
                        x+=': should be resolved ...'
                     else:
-                       x+=': Ok'
+                       # If explicit branch, still add !
+                       branch=xruoa.get('branch','')
+                       checkout=xruoa.get('checkout','')
+                       stable=xruoa.get('stable','')
+                       version=xruoa.get('version','')
+
+                       if branch!='' or checkout!='' or stable!='' or version!='':
+                          xruoa['ignore_pull']='yes'
+                          rp.append(xruoa)
+                          x+=': should be switched to explicit branch ...'
+                       else:
+                          x+=': Ok'
 
                     if o=='con':
                        ck.out(x)
@@ -1761,6 +1888,14 @@ def deps(i):
                  ruoa=xruoa.get('repo_uoa','')
                  ruid=xruoa.get('repo_uid','')
                  rurl=xruoa.get('repo_url','')
+
+                 branch=xruoa.get('branch','')
+                 checkout=xruoa.get('checkout','')
+                 stable=xruoa.get('stable','')
+                 version=xruoa.get('version','')
+
+                 ignore_pull=xruoa.get('ignore_pull','')
+
                  if o=='con':
                     ck.out('')
                     x=''
@@ -1776,6 +1911,11 @@ def deps(i):
                      'data_uoa':ruoa,
                      'current_repos':cr,
                      'url':rurl,
+                     'ignore_pull':ignore_pull,
+                     'branch':branch,
+                     'checkout':checkout,
+                     'stable':stable,
+                     'version':version,
                      'out':o}
                  if ruid!='': ii['data_uid']=ruid
                  if how=='add': ii['gitzip']='yes'
@@ -1929,6 +2069,9 @@ def renew(i):
     """
     Input:  {
               data_uoa                   - data UOA of the repo
+
+              (stable)                   - take stable version (highly experimental)
+              (checkout)                 - checkout (default - stable)
             }
 
     Output: {
@@ -1981,13 +2124,15 @@ def renew(i):
         'module_uoa':work['self_module_uoa'],
         'data_uoa':duoa,
         'data_name':dn,
-        'url':url}
+        'url':url,
+        'stable':i.get('stable',''),
+        'checkout':i.get('checkout','')}
     return ck.access(ii)
 
 ##############################################################################
 # show remote repo in a browser
 
-def show(i):
+def browse(i):
     """
     Input:  {
               data_uoa                   - data UOA of the repo
@@ -2157,3 +2302,268 @@ def form_url(i):
           url+='/ck?'
 
     return {'return':0, 'url':url}
+
+##############################################################################
+# show repositories and their status
+
+def show(i):
+    """
+    Input:  {
+              (data_uoa) - repo UOA
+
+              (reset)    - if 'yes', reset repos
+
+              (stable)   - take stable version (highly experimental)
+              (version)  - checkout version (default - stable)
+            }
+
+    Output: {
+              return       - return code =  0, if successful
+                                         >  0, if error
+              (error)      - error text if return > 0
+            }
+
+    """
+
+    import os
+
+    o=i.get('out','')
+
+    curdir=os.getcwd()
+
+    duoa=i.get('data_uoa','')
+
+    reset=i.get('reset','')
+
+    stable=i.get('stable','')
+    version=i.get('version','')
+    if stable=='yes': version='stable'
+
+    r=ck.list_data({'module_uoa':work['self_module_uoa'], 
+                    'data_uoa':duoa})
+    if r['return']>0: return r
+
+    if o=='con':
+       ck.out('Please wait - it may take some time ...')
+       ck.out('')
+
+    r=ck.reload_repo_cache({}) # Ignore errors
+
+    pp=[]
+    il=0
+    for q in ck.cache_repo_info:
+        # Get repo info
+        qq=ck.cache_repo_info[q]
+
+        d=qq['dict']
+
+        t=d.get('shared','')
+
+        if t!='':
+           duoa=qq['data_uoa']
+
+           if len(duoa)>il: il=len(duoa)
+
+           p=d.get('path','')
+           url=d.get('url','')
+
+           branch=''
+           checkout=''
+
+           if os.path.isdir(p):
+              # Detect status
+              os.chdir(p)
+
+              if reset=='yes':
+                 r=ck.run_and_get_stdout({'cmd':['git','checkout','master']})
+
+              if version!='':
+                 cx=qq.get('dict',{}).get('checkouts',{}).get(version,{})
+                 branch=cx.get('branch','')
+                 checkout=cx.get('checkout','')
+
+                 if branch!='':
+                    r=ck.run_and_get_stdout({'cmd':['git','checkout',branch]})
+
+                 if checkout!='':
+                    r=ck.run_and_get_stdout({'cmd':['git','checkout',checkout]})
+
+              # FGG TBD: we may need to add explicit check for branch/checkout in repo_deps here?
+              # OR MAYBE NOT - need to think ...
+
+              # Get current branch
+              r=ck.run_and_get_stdout({'cmd':['git','rev-parse','--abbrev-ref','HEAD']})
+              if r['return']==0 and r['return_code']==0: 
+                 branch=r['stdout'].strip()
+
+              # Get current checkout
+              r=ck.run_and_get_stdout({'cmd':['git','rev-parse','--short','HEAD']})
+              if r['return']==0 and r['return_code']==0: 
+                 checkout=r['stdout'].strip()
+
+           pp.append({'branch':branch, 'checkout':checkout, 'path':p, 'type':t, 'url':url, 'data_uoa':duoa})
+
+    # Print
+    for q in pp:
+        name=q['data_uoa']
+
+        x=name+' '*(il-len(name))
+
+        branch=q.get('branch','')
+        checkout=q.get('checkout','')
+        url=q.get('url','')
+
+        if branch!='' or checkout!='' or url!='':
+           x+=' ( '+branch+' ; '+checkout+' ; '+url+' )'
+
+        ck.out(x)
+
+    os.chdir(curdir)
+
+    return {'return':0}
+
+##############################################################################
+# reset git repos to default branch and latest checkout (or specific one)
+
+def reset(i):
+    """
+    Input:  {
+              (data_uoa) - repository UOA
+
+              (stable)        - take stable version (highly experimental)
+              (version)       - checkout version (default - stable)
+            }
+
+    Output: {
+              return       - return code =  0, if successful
+                                         >  0, if error
+              (error)      - error text if return > 0
+            }
+
+    """
+
+    i['reset']='yes'
+
+    return show(i)
+
+##############################################################################
+# rename repo
+
+def ren(i):
+    """
+    Input:  {
+              data_uoa  - repo UOA
+
+              (new_data_uoa)
+                 or
+              xcids[0]           - {'data_uoa'} - new data UOA
+            }
+
+    Output: {
+              return       - return code =  0, if successful
+                                         >  0, if error
+              (error)      - error text if return > 0
+            }
+
+    """
+
+    o=i.get('out','')
+
+    duoa=i.get('data_uoa','')
+
+    if duoa=='':
+       return {'return':1, 'error':'repo is not defined'}
+
+    r=ck.access({'action':'load',
+                 'module_uoa':work['self_module_uid'],
+                 'data_uoa':duoa})
+    if r['return']>0: return r
+
+    dd=r['dict']
+    dp=r['path']
+    duoa_real=r['data_uoa']
+    dname=r['data_name']
+
+    nduoa=i.get('new_data_uoa','')
+
+    if nduoa=='':
+       xcids=i.get('xcids',[])
+       if len(xcids)>0: 
+          xcid=xcids[0]
+          nduoa=xcid.get('data_uoa','')
+
+    if nduoa=='':
+       xcids=i.get('cids',[])
+       if len(xcids)>0: 
+          nduoa=xcids[0]
+
+    if nduoa=='': 
+       return {'return':1, 'error':'new repo name is not defined'}
+
+    if nduoa=='': 
+       return {'return':1, 'error':'new repo name is not defined'}
+
+    if nduoa=='local' or nduoa=='default': 
+       return {'return':1, 'error':'new repo name already exists'}
+
+    # Check if such repo doesn't exist
+    r=ck.access({'action':'load',
+                 'module_uoa':work['self_module_uid'],
+                 'data_uoa':nduoa})
+    if r['return']==0:
+       return {'return':1, 'error':'repo already exists'}
+
+    # Update .ckr.json
+    dpp=dd.get('path','')
+    if dpp!='':
+       pckr=os.path.join(dpp,ck.cfg['repo_file'])
+
+       r=ck.load_json_file({'json_file':pckr})
+       if r['return']>0: return r
+
+       dckr=r['dict']
+
+       x=dckr.get('data_uoa','')
+       if x!='' and x==duoa_real: dckr['data_uoa']=nduoa
+
+       x=dckr.get('data_alias','')
+       if x!='' and x==duoa_real: dckr['data_alias']=nduoa
+
+       x=dckr.get('data_name','')
+       if x!='' and x==duoa_real: dckr['data_name']=nduoa
+
+       r=ck.save_json_to_file({'json_file':pckr, 'dict':dckr})
+       if r['return']>0: return r
+
+    # Rename repo entry using internal command
+    r=ck.access({'action':'ren',
+                 'module_uoa':work['self_module_uid'],
+                 'data_uoa':duoa,
+                 'new_data_uoa':nduoa,
+                 'common_func':'yes'})
+    if r['return']>0: return r
+
+    # Recache repos
+    r1=recache({'out':o})
+    if r1['return']>0: return r1
+
+    return r
+
+##############################################################################
+# rename repo
+
+def rename(i):
+    """
+    Input:  {
+               See "ck ren repo --help"
+            }
+
+    Output: {
+              return       - return code =  0, if successful
+                                         >  0, if error
+              (error)      - error text if return > 0
+            }
+
+    """
+
+    return ren(i)
