@@ -622,8 +622,10 @@ def start(i):
 
               (browser)     - if 'yes', open browser
               (template)    - if !='', add template
-              (cid)         - view a given entry
-              (extra_url)   - extra text for URL
+              (wcid)         - view a given entry
+                or
+              (cid)
+              (extra_url)   - extra URL
             }
 
     Output: {
@@ -669,7 +671,10 @@ def start(i):
        if i.get('template','')!='':
           ext='template='+i['template']
 
-       cid=i.get('cid','')
+       cid=i.get('wcid','')
+       if cid=='':
+          cid=i.get('cid','')
+
        if cid!='' and cid!='web':
           if ext!='': ext+='&'
           ext+='wcid='+cid
